/* LCD Include-Datei
 *
 * Spezielle Version fr 4x20 LCD 
 *
 * LCD_CTRL stellt ein, welcher Controllertyp (wichtig fr DD-RAM-Adressen)
 * Einige Routinen sind fr ein EA DIP204-4 mit KS0073 Controller notwendig, aber auskommentiert
 *
 * Das LCD ist folgendermaen angeschlossen:
 * PC4     = RS
 * PC5     = Enable
 * PC0-PC3 = D4-D7
 * R/W ist n.c.
 *
 *
 */

#include <avr/io.h>
#include <util/delay.h> 
#include <math.h>         
#include "lcd_tools.h"

#define LCD_CTRL 0 		// (0 = HD44780, 1= KS0073)
#define delay(us)  _delay_loop_2 (((F_CPU/4000)*us)/1000)		// wartet s

/**
	@brief	Speichert neue Zeichen im LCD: Werkzeugschlssel und Ausrufezeichen + Grozahlen
			Mu nur einmal aufgerufen werden
	@param 	none
	@return	none
*/
void lcd_new_char ()
{

//----------------------------------------------------------
// LCD-Zeichen 1 (Punkt)
//----------------------------------------------------------
	lcd_write (0x48, 0);	
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (4, 1);
	lcd_write (14, 1);
	lcd_write (14, 1);
	lcd_write (4, 1); 

//----------------------------------------------------------
// LCD-Zeichen 2
//----------------------------------------------------------
	lcd_write (0x50, 0);	
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1); 

//----------------------------------------------------------
// LCD-Zeichen 3
//----------------------------------------------------------
	lcd_write (0x58, 0);	
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1); 

//----------------------------------------------------------
// LCD-Zeichen 4
//----------------------------------------------------------
	lcd_write (0x60, 0);	
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (30, 1);
	lcd_write (28, 1);
	lcd_write (24, 1);
	lcd_write (16, 1); 

//----------------------------------------------------------
// LCD-Zeichen 5
//----------------------------------------------------------
	lcd_write (0x68, 0);	
	lcd_write (16, 1);
	lcd_write (24, 1);
	lcd_write (28, 1);
	lcd_write (30, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1); 

//----------------------------------------------------------
// LCD-Zeichen 6
//----------------------------------------------------------
	lcd_write (0x70, 0);	
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (31, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1);
	lcd_write (0, 1); 


	lcd_cls();
}

// Runden. Ihr bergebt als ersten Parameter die zu rundende Zahl und als zweiten Parameter die Anzahl von Nachkommastellen.
// Nach der Operation double r=Round(1.34556,2) steht in r also 1.35
double newRound(double amount, int power_of_ten)
{
	double pten; /* 10 hoch power_of_ten */
	double intpart; /* Integerteil der durch pten geteilten Zahl */
	double halfsign; /* Plus oder Minus 0.5 fr korrekte Rundung */
	double temp;
	temp = (double)power_of_ten;
	
	pten= pow((double)(10), temp); /* auf Zehnerpotenz definieren */
	halfsign = (amount < 0.0) ? - 0.5 : 0.5;
	modf( (double)(amount * pten + halfsign), &intpart);
	return ((double)(intpart / pten) );
}


/**
	@brief	Enable Leitung togglen
	@param 	none
	@return	none
*/
void lcd_flash_e (void)
{
	PORTC = PORTC | ( 1<<DDC5 ) ;		// Enable = HIGH
	delay(100);							// kurz warten, damit Flanke entsteht
	PORTC = PORTC & ~( 1<<DDC5 ) ;		// Enable = LOW
	delay(100);							// Warten, bis der durch Enable angelegte Befehl hoffentlich verarbeitet wurde
}

/**
	@brief	Kommando oder Zeichen an Display senden: 
	@param 	rs=0 => Kommando, rs=1 => Zeichen
	@return	none
*/
void lcd_write (uint8_t data, uint8_t rs)   
{
	uint8_t dataBits ;
  
    if (rs)    				// write data (RS=1, RW=0) 
       dataBits=0x10;		// RS liegt an Pin 4 = B 0001 0000 = H 10
    else     				// write instruction (RS=0, RW=0) 
       dataBits=0; 			
	
    PORTC = dataBits | (data>>4);		// output high nibble first, zzgl. Zustand fr RS-Leitung
	lcd_flash_e ();    

	PORTC = dataBits | (data&0x0F);		// output low nibble, zzgl. RS-Leitung
	lcd_flash_e ();
}

/**
	@brief	Display loeschen
	@param 	none
	@return	none
*/
void lcd_cls (void)   
{
	lcd_write(0x00,0);   	// B 0000 0000 => Clear
	delay(4000);			// dauert eine Weile
	lcd_write(0x01,0);   	// B 0000 0001 => Cursor Home
	delay(4000);			// dauert eine Weile, Wert ausprobiert
}

/**
	@brief	Zeichenausgabe
	@param 	none
	@return	none
*/
void lcd_writechar ( char zeichen)   
{
  lcd_write (zeichen, 1);   
}

/**
	@brief	gibt eine Zeichenkette auf dem LCD aus
	@param 	none
	@return	none
*/
void lcd_writetext ( char *text)    
{
  uint8_t i = 0;
  while (text[i]!=0) 
  {
    lcd_writechar(text[i]);
    i++;
  }
}

/**
	@brief	Zeilenwechsel
	@param 	none
	@return	none
*/
void lcd_gotoline (uint8_t zeile)    
{
	if (LCD_CTRL == 0)
	{
		if (zeile == 1) lcd_write(0x80,0);   		// B 1000 0000 => DD-RAM Adress Set 1. Zeile/1.Spalte 
		if (zeile == 2) lcd_write(0x80+0x40,0);   	// B 1010 0000 => DD-RAM Adress Set 2. Zeile/1.Spalte 
		if (zeile == 3) lcd_write(0x80+0x14,0);   	// B 1010 0000 => DD-RAM Adress Set 3. Zeile/1.Spalte 
		if (zeile == 4) lcd_write(0x80+0x54,0);   	// B 1010 0000 => DD-RAM Adress Set 4. Zeile/1.Spalte 
	}
	else
	{
		if (zeile == 1) lcd_write(0x80,0);   		// B 1000 0000 => DD-RAM Adress Set 1. Zeile/1.Spalte 
		if (zeile == 2) lcd_write(0x80+0x20,0);   	// B 1010 0000 => DD-RAM Adress Set 2. Zeile/1.Spalte 
		if (zeile == 3) lcd_write(0x80+0x40,0);   	// B 1010 0000 => DD-RAM Adress Set 3. Zeile/1.Spalte 
		if (zeile == 4) lcd_write(0x80+0x60,0);   	// B 1010 0000 => DD-RAM Adress Set 4. Zeile/1.Spalte 
	}

}

/**
	@brief	Cursorpositionieren 
	@param 	Zeile, Spalte
	@return	none
*/
void lcd_gotopos (uint8_t zeile, uint8_t spalte)    
{
	if (LCD_CTRL == 0)
	{
		if (zeile == 1) lcd_write(0x80+0x00+spalte-1,0);   // DD-RAM Adress 1. Zeile + Spalte
		if (zeile == 2) lcd_write(0x80+0x40+spalte-1,0);   // DD-RAM Adress 2. Zeile + Spalte 
		if (zeile == 3) lcd_write(0x80+0x14+spalte-1,0);   // DD-RAM Adress 3. Zeile + Spalte 
		if (zeile == 4) lcd_write(0x80+0x54+spalte-1,0);   // DD-RAM Adress 4. Zeile + Spalte 
	}
	else
	{
		if (zeile == 1) lcd_write(0x80+0x00+spalte-1,0);   // DD-RAM Adress 1. Zeile + Spalte
		if (zeile == 2) lcd_write(0x80+0x20+spalte-1,0);   // DD-RAM Adress 2. Zeile + Spalte 
		if (zeile == 3) lcd_write(0x80+0x40+spalte-1,0);   // DD-RAM Adress 3. Zeile + Spalte 
		if (zeile == 4) lcd_write(0x80+0x60+spalte-1,0);   // DD-RAM Adress 4. Zeile + Spalte 
	}
}

/**
	@brief	Gibt eine einstellige (0-9) Zahl in groschreibweise (3 Zeilen, 3 Spalten) in Zeile 2-4 aus
	@param 	Zahl
			Zeichenspalte
	@return	none
*/
void lcd_bignum (uint8_t zahl, uint8_t spalte)
{
	lcd_gotopos (2,spalte);
	switch (zahl)
	{
		case 0: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3); break;  
		case 1: lcd_writechar(3);lcd_writechar(3);lcd_writechar(32) ; break;
		case 2: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 3: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 4: lcd_writechar(3);lcd_writechar(32);lcd_writechar(32); break;
		case 5: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 6: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 7: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 8: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
		case 9: lcd_writechar(3);lcd_writechar(3);lcd_writechar(3)  ; break;
	}

	lcd_gotopos (3,spalte);
	switch (zahl)
	{
		case 0: lcd_writechar(2);lcd_writechar(32);lcd_writechar(2) ; break;
		case 1: lcd_writechar(32);lcd_writechar(2);lcd_writechar(32)  ; break;
		case 2: lcd_writechar(3);lcd_writechar(3);lcd_writechar(4)      ; break;
		case 3: lcd_writechar(32);lcd_writechar(3);lcd_writechar(4)     ; break;
		case 4: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
		case 5: lcd_writechar(2);lcd_writechar(3);lcd_writechar(3)    ; break;
		case 6: lcd_writechar(2);lcd_writechar(3);lcd_writechar(3)    ; break;
		case 7: lcd_writechar(32);lcd_writechar(3);lcd_writechar(4)     ; break;
		case 8: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
		case 9: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
	}
	
	lcd_gotopos (4,spalte);
	switch (zahl)
	{
		case 0: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
		case 1: lcd_writechar(32);lcd_writechar(2);lcd_writechar(32)   ; break;
		case 2: lcd_writechar(2);lcd_writechar(3);lcd_writechar(3)   ; break;
		case 3: lcd_writechar(3);lcd_writechar(3);lcd_writechar(5)      ; break;		
		case 4: lcd_writechar(32);lcd_writechar(32);lcd_writechar(2)   ; break;
		case 5: lcd_writechar(3);lcd_writechar(3);lcd_writechar(2)   ; break;
		case 6: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
		case 7: lcd_writechar(32);lcd_writechar(2);lcd_writechar(32)   ; break;
		case 8: lcd_writechar(2);lcd_writechar(3);lcd_writechar(2)  ; break;
		case 9: lcd_writechar(3);lcd_writechar(3);lcd_writechar(2)   ; break;
	}
	
}

/**
	@brief	Public
			Gibt eine max. fnstellige (inkl. Komma) Zahl in groschreibweise (3 Zeilen, 3 Spalten) 
			rechtsformatiert ohne fhrende Nullstellen aus
	@param 	Zahl (-9.999 bis 99.999
			Anzahl der darzustellenden Nachkommastellen, 0=keine
	@return	none
*/
void lcd_write_bignums (double zahl, uint8_t anz_nachkomma)
{
	lcd_ausgabe_bignums (zahl, anz_nachkomma, 0);
}

/**
	@brief	Private
			Gibt eine max. fnstellige (inkl. Komma) Zahl in groschreibweise (3 Zeilen, 3 Spalten) 
			rechtsformatiert ohne fhrende Nullstellen aus
	@param 	Zahl (-9.999 bis 99.999
			Anzahl der darzustellenden Nachkommastellen, 0=keine
			Anzahl Stellen (fhrende Nullen) bei Nachkommadarstellung
	@return	none
*/
void lcd_ausgabe_bignums (double zahl, uint8_t anz_nachkomma, uint8_t stellen_anz)
{
	int32_t output=0,
			output_nachkomma=0;
	uint8_t digits=0,
			pos=2,		// Anfangsspalte fr Ausgabe (fngt nicht bei 1 an, damit zahlen am rechten Rand stehen)
			i,u;
	double	vorkomma,nachkomma;
	
	if (stellen_anz == 0)
	{
		zahl = newRound (zahl, anz_nachkomma);		// Zahl auf tatschlche Nachkommastellenzahl runden (9,99 -> 10,0 bei 1. Nachkommastelle)
		
		nachkomma = modf (zahl, &vorkomma);			// Zerlege in Nachkomma und Vorkommateil (12,34 -> 12 und 0,34)
		if (nachkomma < 0)
			nachkomma = nachkomma * -1;		// positive Zahl
		output = (int32_t)vorkomma;
		output_nachkomma = (int32_t)(nachkomma * pow ((double)(10), anz_nachkomma));
	}
	else
		output = zahl;

	if (zahl < 0)
		output = (int32_t)output * -1;	// positive Zahl
	
	if (stellen_anz == 0)	// wenn keine Angabe zur Anzahl der Stellen (Aufruf durch Benutzer)
	{
		// Anzahl der Stellen ermitteln
		if (output >= 0) digits=1;
		if (output >= 10) digits=2;
		if (output >= 100) digits=3;
		if (output >= 1000) digits=4;
		if (output >= 10000) digits=5;
		if (output >= 100000) digits=6;		// Abbruchkriterium
		
		if (zahl < 0) digits++;			// Minuszeichen
		if (anz_nachkomma > 0)
			digits += anz_nachkomma + 1;	// Nachkommastellen + Kommazeichen
	}
	else	// Anzahl der Stellen ergibt sich aus Nachkommaanteil (Rekursiver Aufruf)
		digits = stellen_anz;
		
	if (digits <= 5)	// Ausgabe nur wenn max 5 Stellen
	{
		pos += (5-digits)*4;
		if (anz_nachkomma > 0)
			pos += 2;			// Korrekturwert fr Spaltanzahl bei Nachkommastellen fr das Kommazeichen (nur 1 Breit statt der blichen 3)
		
		if (stellen_anz == 0)	// wenn Vorkommaanteil ausgegeben wird.
		{
			// Lschen von Zeichen, die aus der vorherigen Ausgabe noch da stehen, wenn die Zahlen kleiner werden (bspw. 3-stellig ->2-stellig)
			for (i=2; i <= 4; i++)			// jede Zeile
			{
				lcd_gotopos (i,2);	
				for (u=2; u < pos; u++)		// jede Spalte bis erste Spalte des neuen Zeichens 
					lcd_writechar (' ');	// Leerzeichen
			}
		}
		
		// negatives Vorzeichen ausgeben
		if (zahl < 0)
		{
			lcd_gotopos (2, pos); lcd_writechar(32);lcd_writechar(32);lcd_writechar(32);
			lcd_gotopos (3, pos); lcd_writechar(3);lcd_writechar(3);lcd_writechar(3); 
			lcd_gotopos (4, pos); lcd_writechar(32);lcd_writechar(32);lcd_writechar(32);
			zahl *= -1.0;		// positive Zahl
			pos +=4;
		}
	
		// Zehntausender-Dezimalstelle oder Leerzeichen (Ausblenden fuehrender Nullen)
		if (zahl >= 10000)
		{
			lcd_bignum (output/10000, pos); 
			pos += 4;
		}
		else if (stellen_anz >= 5)
		{
			lcd_bignum (0, pos); 
			pos += 4;
		}
		output = output % 10000;
		
		// Tausender-Dezimalstelle oder Leerzeichen
		if (zahl >= 1000)
		{
			lcd_bignum (output/1000, pos); 
			pos += 4;
		}
		else if (stellen_anz >= 4)
		{
			lcd_bignum (0, pos); 
			pos += 4;
		}
		output = output % 1000;
		
		// Hunderter-Dezimalstelle oder Leerzeichen
		if (zahl >= 100)
		{
			lcd_bignum (output/100, pos); 
			pos += 4;
		}
		else if (stellen_anz >= 3)
		{
			lcd_bignum (0, pos); 
			pos += 4;
		}
		output = output % 100;
		
		// Zehner-Dezimalstelle oder Leerzeichen
		if (zahl >= 10)
		{
			lcd_bignum (output/10, pos); 
			pos += 4;
		}
		else if (stellen_anz >= 2)
		{
			lcd_bignum (0, pos); 
			pos += 4;
		}
		output = output % 10;
		
		// einer Dezimalstelle 
		lcd_bignum (output, pos); 
		
		if (anz_nachkomma > 0)	// Nachkommastellen durch Rekursion ausgeben
		{
			pos +=3;	// an die Leerstelle nach der letzten Zahl gehen (3 statt 4) und 3x3 Blanks ausgeben. Ist wichtig, wenn Dezimalpunkt wandert bei untersch. Anz. Nachkommast.
			lcd_gotopos (2, pos); lcd_writechar(32);lcd_writechar(32);lcd_writechar(32);
			lcd_gotopos (3, pos); lcd_writechar(32);lcd_writechar(32);lcd_writechar(32);
			lcd_gotopos (4, pos); lcd_writechar(32);lcd_writechar(32);lcd_writechar(32);
			lcd_gotopos (4, pos+1);
			lcd_writechar(1);		// Dezimalpunkt (Komma) ausgeben
			lcd_ausgabe_bignums (output_nachkomma, 0, anz_nachkomma);
		}

	}
}

/**
	@brief	gibt eine int-Zahl auf dem Display aus. Integriert, damit auf string.h verzichtet werden kann
	@param 	Zahl
			Anzahl der Stellen (fhrende Leerzeichen)
			space=0 => fhrende Nullen, 1 => fhrende Leerzeichen
	@return	none
*/
void lcd_writezahl (int32_t zahl, uint8_t pos, uint8_t spacer)       
{
   int32_t output;
   char sign = '0';
  
  pos--;       // Korrektur des Parameters fr Benutzerfreundlichkeit
  
  if (spacer) sign=' ';

  // negatives Vorzeichen oder Leerzeichen
  if (zahl < 0)
  {
	lcd_writechar ('-'); 
	zahl *= -1;		// positive Zahl
  }

  output = zahl;
  
  // Hunderttausender-Dezimalstelle oder Leerzeichen (Ausblenden fuehrender Nullen)
  if (zahl >= 100000) lcd_writechar (0x30 + output/100000); else if (pos >= 5) lcd_writechar (sign);
  output = output % 100000;

  // Zehntausender-Dezimalstelle oder Leerzeichen (Ausblenden fuehrender Nullen)
  if (zahl >= 10000) lcd_writechar (0x30 + output/10000); else if (pos >= 4) lcd_writechar (sign);
  output = output % 10000;

  // Tausender-Dezimalstelle oder Leerzeichen
  if (zahl >= 1000) lcd_writechar (0x30 + output/1000); else if (pos >= 3) lcd_writechar (sign);
  output = output % 1000;

  // Hunderter-Dezimalstelle oder Leerzeichen
  if (zahl >= 100) lcd_writechar (0x30 + output/100); else if (pos >= 2) lcd_writechar (sign);
  output = output % 100;

  // Zehner-Dezimalstelle oder Leerzeichen
  if (zahl >= 10) lcd_writechar (0x30 + output/10); else if (pos >= 1) lcd_writechar (sign);

  // einer Dezimalstelle oder Leerzeichen
  lcd_writechar (0x30 + output % 10);
}

/**
	@brief	Zuerst den 8-Bit-Modus aktivieren, Ist wichtig, falls LCD schon im 4-Bit Modus war und dann 
			nach einem Programm-Reset vergeblich versucht wrde, den 4-Bit Modus erneut zu aktivieren.
			Dann in 4-Bit Modus umschalten.
	@param 	none
	@return	none
*/
void lcd_set4bit (void)
{
    PORTC = 0x3;			// 8-Bit Modus aktivieren (B 0011 0000 = H 30, High Nibble => B 0011 = H 03
	lcd_flash_e ();    		// Datenbernahme
	delay(200);			// sicherheitshalber warten da wir BUSY nicht berprfen knnen 

	// Das ganze 2x wiederholen:
	PORTC = 0x3;		
	lcd_flash_e ();    
	delay(200);			
	
	PORTC = 0x3;		
	lcd_flash_e ();    
	delay(200);				

	lcd_write(0x02,0);		// B 0000 0010 => in 4-Bit-Modus umschalten 
	delay(200);				
}

/**
	@brief	Portrichtung einstellen. Extern, damit man dies separat aufrufen kann, wenn bspw. LCD schon initialisiert ist und
			per Watchdog ein Reset ausgelst wurde, man dann aber nicht das Display neu initialisieren mchte, damit es schneller weiter geht
	@param 	none
	@return	none
*/
void lcd_port_ini (void)
{
	DDRC = 0x3F;		// setze Portrichtung (1 = Ausgang): 0011 1111 
}


/**
	@brief	Schaltet das Display ein und auch die Hintergrundbeleuchtung ber PD7
	@param 	none
	@return	none
*/
void lcd_on (void)
{
	lcd_write(0x0C,0);		// B 0000 1100 => Display On, Cursor aus, Blinken aus
	DDRD |= (1 << DDD7);	// PD7 als Ausgang 
	PORTD |= (1 << PD7);	// PD7 ein
}

/**
	@brief	Schaltet das Display aus und auch die Hintergrundbeleuchtung ber PD7
	@param 	none
	@return	none
*/
void lcd_off (void)
{
	lcd_write(0x08,0);		// B 0000 1100 => Display Off, Cursor aus, Blinken aus
	DDRD |= (1 << DDD7);	// PD7 als Ausgang 
	PORTD &= ~(1 << PD7);	// PD7 aus	
}

/**
	@brief	Display initialisieren. Einmal als erstes aufrufen
	@param 	none
	@return	none
*/
void lcd_ini (void)
{
	lcd_port_ini();
	DDRC = 0x3F;			// setze Portrichtung (1 = Ausgang): 0011 1111 

	delay(20000);			// 20ms warten bis LCD wirklich bereit! Wir knnen nicht BUSY auswerten. 
							// Es gibt Displays die so lange brauchen (z. B. Optrex DMC20481)

	lcd_set4bit();			// ab jetzt 4-Bit-Modus
	
	lcd_write(0x2C,0);		// 4x20 Spezial: Function Set (0010 1100): 4-Bit, 2 Line, RE-Bit, Dot Scroll, Normal Mode
	lcd_write(0x09,0);		// 4x20: Extended Function Set (0000 1001): 5 dot font, Normal Cursor, 4 Line Display

	lcd_write(0x28,0);		// B 0010 1000 => Function Set: 4Bit (kein 8-Bit Bus), zweizeiliges Display, 5x7 Dots/Zeichen (kein 5x10), RE-Bit aus
//	lcd_write(0x0F,0);		// B 0000 1100 => Display On/Off: Display ein, Cursor an, Blinken an
	lcd_on();
	lcd_write(0x06,0);		// B 0000 0110 => Entry Mode Set: DD-RAM autom. inkr., Cursor bewegen
	
	lcd_cls();				// Display lschen
}


