/* Demonstriert den 16-Bit Timer 1
 *
 * Input Capture ICP1
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>				// dtostrf
#include "lcd_tools.h"

volatile uint16_t counter = 0;		// Zaehler fuer Ueberlaeufe
volatile double time = 0;			// Millisekunden
volatile uint8_t running = 0;		// Zustand der Uhr
volatile uint8_t sec = 0;			// Anzahl Sekunden

// Timer/Counter1 Capture Event 
ISR (TIMER1_CAPT_vect)
{
	if (running)			// wenn Uhr laeuft
	{
		counter = ICR1;					// Zaehlerstand retten
		time = counter/(F_CPU/256.0);	// = Millisekunden
		running = 0;					// Uhr steht
	}
	else	// Uhr war gestopt, nun starten
	{
		TCNT1 = 0;		// zurueksetzen des Counters
		running = 1;	// Uhr laeuft
	}
}

// Timer/Counter1 Compare Match
ISR (TIMER1_COMPA_vect)		// je Sekunde einmal
{
	if (running)	// wenn Uhr laeuft
	{
		sec++;		// Sekunden zaehlen
		lcd_gotopos (2, 7);
		lcd_writezahl (sec, 3, 1);
	}
}

int main()
{
	uint8_t text[10];
	
	lcd_ini();
	lcd_writetext ("  *** STOPUHR ***");
	
	TIMSK |= (1 << TICIE1) | (1 << OCIE1A);// Input Capture Interrupt Enable+Output Compare A Match Interrupt Enable
	OCR1A = (uint16_t)(F_CPU/256);			// Vergleichswert 1 Sekunde
	sei();									// IRQs enable
	TCCR1B |= (1 << WGM12);				// CTC Modus
	TCCR1B |= (1 << ICNC1) | (1 << ICES1);	// Noise Canceler + steigende Flanke an ICP1
	TCCR1B |= (1 << CS12);	// Prescaler: 1/256

	while (1)		// endlos
	{
		if (running == 1)	// wenn Uhr laeuft und dies hier 
		{					// noch nicht aufgerufen wurde
			lcd_gotoline (2);
			lcd_writetext ("Zeit:   0           ");
			running = 2;	// Damit dieser Teil nur einmal aufgerufen wird
		}
		if (counter)		// Stop gedrueckt => Wert in Variable
		{
			dtostrf (sec+time, 7, 3, text);
			lcd_gotoline (2);
			lcd_writetext ("Zeit: ");
			lcd_writetext (text);
			lcd_writetext (" Sek.");
			counter = 0;
			sec=0;
		}
	}
	return 0;
}
